open Base

type label = string 

type register = 
  | A0 | A1 | V0 | S0 | RA | SP | FP | GP | T0 | T1

type address =
  | Alab of string
  | Areg of int * register

type operand =
  | Oimm of int
  | Oreg of register


type instruction =
  | Move of register * register
  | Li of register * int
  | La of register * string
  | Lw of register * address
  | Sw of register * address
  | Bin of op * register * register * operand
  | B of label
  | Beqz of register * label
  | Bnez of register * label
  | Blt of register * register * label
  | Ble of register * register * label
  | Jal of string
  | Jr of register
  | Syscall
  | Label of string

type data = 
  | Asciiz of string * string
  | Word of string * int

(* Data structure to avoid list append complexity *)

type code = 
  | List of instruction list
  | Conc of code * code

let nop = List []

let mips l = List l

let (++) c1 c2 = match c1, c2 with
  | List [] , c | c, List [] -> c
  | _ -> Conc (c1, c2)

let rec code_iter f = function 
    List l -> List.iter f l
  | Conc (z1,z2) -> code_iter f z1; code_iter f z2

type program = {
  text : code;
  data : data list;
}

open Format

let print_register fmt = function
  | A0 -> pp_print_string fmt "$a0"
  | A1 -> pp_print_string fmt "$a1"
  | V0 -> pp_print_string fmt "$v0"
  | S0 -> pp_print_string fmt "$s0"
  | RA -> pp_print_string fmt "$ra"
  | SP -> pp_print_string fmt "$sp"
  | FP -> pp_print_string fmt "$fp"
  | GP -> pp_print_string fmt "$gp"
  | T0 -> pp_print_string fmt "$t0"
  | T1 -> pp_print_string fmt "$t1"

let print_op fmt = function
  | Plus -> pp_print_string fmt "add"
  | Minus -> pp_print_string fmt "sub"
  | Mult -> pp_print_string fmt "mul"
  | Div -> pp_print_string fmt "div"
  | Eq -> pp_print_string fmt "seq"
  | Lt -> pp_print_string fmt "slt"
  | Leq -> pp_print_string fmt "sle"
  | Gt -> pp_print_string fmt "sgt"
  | Geq -> pp_print_string fmt "sge"

  

let print_address fmt = function
  | Alab s -> pp_print_string fmt s
  | Areg (ofs, r) -> fprintf fmt "%d(%a)" ofs print_register r

let print_operand fmt = function
  | Oimm i -> pp_print_int fmt i
  | Oreg r -> print_register fmt r

let print_instruction fmt = function
  | Move (dst, src) -> 
      fprintf fmt "\tmove %a, %a\n" print_register dst print_register src
  | Li (r, i) ->
      fprintf fmt "\tli   %a, %d\n" print_register r i
  | La (r, s) ->
      fprintf fmt "\tla   %a, %s\n" print_register r s
  | Lw (r, a) ->
      fprintf fmt "\tlw   %a, %a\n" print_register r print_address a
  | Sw (r, a) ->
      fprintf fmt "\tsw   %a, %a\n" print_register r print_address a
  | Bin (bop, dst, src, op) ->
      fprintf fmt "\t%a  %a, %a, %a\n" 
	print_op bop print_register dst print_register src print_operand op
  | B l ->
      fprintf fmt "\tb    %s\n"  l
  | Beqz (r, l) ->
      fprintf fmt "\tbeqz %a, %s\n" print_register r  l
  | Bnez (r, l) ->
      fprintf fmt "\tbnez %a, %s\n" print_register r  l
  | Blt (r1,r2,l)  ->
      fprintf fmt "\tblt  %a, %a, %s\n" 
	print_register r1 print_register r2  l
  | Ble (r1,r2,l)  ->
      fprintf fmt "\tble  %a, %a, %s\n" 
	print_register r1 print_register r2  l

  | Jal s ->
      fprintf fmt "\tjal  %s\n" s
  | Jr r ->
      fprintf fmt "\tjr   %a\n" print_register r
  | Syscall ->
      fprintf fmt "\tsyscall\n"
  | Label s ->
      fprintf fmt "%s:\n" s

let print_data fmt = function
  | Asciiz (l, s) -> 
      fprintf fmt "%s:\t.asciiz %S\n" l s
  | Word (l, n) ->
      fprintf fmt "%s:\t.word %d\n" l n

let print_program fmt p =
  fprintf fmt "\t.text\n";
  code_iter (print_instruction fmt) p.text;
  fprintf fmt "\t.data\n";
  List.iter (print_data fmt) p.data

let pushr r = mips [Bin(Minus,SP,SP,Oimm(4));Sw(r,Areg(0,SP))]
let popr r = mips [Lw(r,Areg(0,SP));Bin(Plus,SP,SP,Oimm(4))]


let printr r = mips
    [Move(A0,r); Li (V0,1); Syscall; Li (V0,11); Li (A0,10); Syscall]

let printa0 = mips
    [Li (V0,1); Syscall; Li (V0,11); Li (A0,10); Syscall]

let pushg r x = mips
    [Lw(r,Alab(x))]

